@@ -47,6 +47,18 @@ import re
 import ssl
 from io import open
 
+if False:
+    from typing import Optional
+    from typing import Dict
+
+# Work around with a striptime race we see every now and then with docker monitor run() method.
+# That race would occur very rarely, since it depends on the order threads are started and when
+# strptime is first called.
+# See:
+# 1. https://github.com/scalyr/scalyr-agent-2/pull/700#issuecomment-761676613
+# 2. https://bugs.python.org/issue7980
+import _strptime  # NOQA
+
 try:
     from __scalyr__ import SCALYR_VERSION
     from __scalyr__ import scalyr_init
@@ -65,6 +77,11 @@ scalyr_init()
 
 import six
 
+try:
+    import glob
+except ImportError:
+    import glob2 as glob  # type: ignore
+
 import scalyr_agent.scalyr_logging as scalyr_logging
 import scalyr_agent.util as scalyr_util
 import scalyr_agent.remote_shell as remote_shell
@@ -114,7 +131,7 @@ AGENT_LOG_FILENAME = "agent.log"
 AGENT_NOT_RUNNING_MESSAGE = "The agent does not appear to be running."
 
 # Message which is logged when locale used for the scalyr agent process is not UTF-8
-NON_UTF8_LOCALE_WARNING_MESSAGE = """
+NON_UTF8_LOCALE_WARNING_LINUX_MESSAGE = """
 Detected a non UTF-8 locale (%s) being used. You are strongly encouraged to set the locale /
 coding for the agent process to UTF-8. Otherwise things won't work when trying to monitor files
 with non-ascii content or non-ascii characters in the log file names. On Linux you can do that by
@@ -123,6 +140,15 @@ setting LANG and LC_ALL environment variable: e.g. export LC_ALL=en_US.UTF-8.
     "\n", " "
 )
 
+NON_UTF8_LOCALE_WARNING_WINDOWS_MESSAGE = """
+Detected a non UTF-8 locale (%s) being used. You are strongly encouraged to set the locale /
+coding for the agent process to UTF-8. Otherwise things won't work when trying to monitor files
+with non-ascii content or non-ascii characters in the log file names. On Windows you can do that by
+setting setting PYTHONUTF8=1 environment variable.
+""".strip().replace(
+    "\n", " "
+)
+
 # Work around for a potential race which may happen when threads try to resolve a unicode hostname
 # or similar
 # See:
@@ -302,8 +328,13 @@ class ScalyrAgent(object):
                 config_file_path, log_warnings=log_warnings
             )
 
-            # check if not a tty and override the no check remote variable
-            if not sys.stdout.isatty():
+            # NOTE: isatty won't be available on Redirector object on Windows when doing permission
+            # escalation so we need to handle this scenario as well
+            isatty_func = getattr(getattr(sys, "stdout", None), "isatty", None)
+            if not isatty_func or (isatty_func and not isatty_func()):
+                # If stdout.atty is not available or if it is and returns False, we fall back to
+                # "check_remote_if_no_tty" config option
+                # check if not a tty and override the no check remote variable
                 no_check_remote = not self.__config.check_remote_if_no_tty
         except Exception as e:
             # We ignore a bad configuration file for 'stop' and 'status' because sometimes you do just accidentally
@@ -985,6 +1016,61 @@ class ScalyrAgent(object):
                 % (raw_scalyr_server, scalyr_server)
             )
 
+    def __get_log_files_initial_positions(self, only_new_files=False):
+        # type: (bool) -> Optional[Dict]
+        """
+        According to the current configuration, determine all agent log files (including main agent.log and
+        multi-process worker sessions logs), create those files (if they do not exist) and get their current positions.
+
+        The preliminarily creation of the log files, in case of their absence, is needed because agent log messages may
+        be written to the agent log file before this log file is processed by the copying manager. In this case,
+        the copying manager will start sending log file from its current position, skipping everything before that and
+        causing data loss.
+
+        :param only_new_files: This option should be used when configuration is reloaded and more worker sessions were
+        added. We ignore existing files and only process log files for new worker sessions.
+        :return:
+        """
+
+        # all paths for all agent log files.
+        log_file_paths_to_check = [self.__log_file_path]
+
+        # we also add log files for all worker session if they are in multiprocess configuration.
+        if self.__config.use_multiprocess_workers:
+            for worker_session_id in self.__config.get_session_ids_from_all_workers():
+                log_file_path = self.__config.get_worker_session_agent_log_path(
+                    worker_session_id
+                )
+                log_file_paths_to_check.append(log_file_path)
+
+        log_file_paths = []
+        for log_file_path in log_file_paths_to_check:
+            include_file = True
+            if os.path.isfile(log_file_path):
+                if only_new_files:
+                    # Only new files are handled, skip this file.
+                    include_file = False
+            else:
+                # the file does not exist, create it now so we can get its current position.
+                with open(log_file_path, "w"):
+                    pass
+
+            if include_file:
+                log_file_paths.append(log_file_path)
+
+        logs_initial_positions = {}
+
+        # get initial positions for the files. The copying manager will start copying files from those positions.
+        for log_path in log_file_paths:
+            log_position = self.__get_file_initial_position(log_path)
+            if log_position is not None:
+                logs_initial_positions[log_path] = log_position
+
+        if logs_initial_positions:
+            return logs_initial_positions
+        else:
+            return None
+
     def __run(self, controller):
         """Runs the Scalyr Agent 2.
 
@@ -1037,16 +1123,6 @@ class ScalyrAgent(object):
 
                 self.__update_debug_log_level(self.__config.debug_level)
 
-                # We record where the log file currently is so that we can (in the worse case) start copying it
-                # from this position.  That way we capture the first 'Starting scalyr agent' call.
-                agent_log_position = self.__get_file_initial_position(
-                    self.__log_file_path
-                )
-                if agent_log_position is not None:
-                    logs_initial_positions = {self.__log_file_path: agent_log_position}
-                else:
-                    logs_initial_positions = None
-
                 start_up_msg = scalyr_util.get_agent_start_up_message()
                 log.info(start_up_msg)
                 log.log(scalyr_logging.DEBUG_LEVEL_1, start_up_msg)
@@ -1056,7 +1132,10 @@ class ScalyrAgent(object):
                 # content
                 _, encoding, _ = scalyr_util.get_language_code_coding_and_locale()
                 if encoding.lower() not in ["utf-8", "utf8"]:
-                    log.warn(NON_UTF8_LOCALE_WARNING_MESSAGE % (encoding))
+                    if sys.platform.startswith("win"):
+                        log.warn(NON_UTF8_LOCALE_WARNING_WINDOWS_MESSAGE % (encoding))
+                    else:
+                        log.warn(NON_UTF8_LOCALE_WARNING_LINUX_MESSAGE % (encoding))
 
                 self.__controller.emit_init_log(log, self.__config.debug_init)
 
@@ -1084,11 +1163,17 @@ class ScalyrAgent(object):
                     wt.start(logs_initial_positions)
                     return wt, wt.copying_manager, wt.monitors_manager
 
+                # We record where the log files(including multiprocess worker sessions logs) currently are
+                # so that we can (in the worse case) start copying them from those position.
+                logs_initial_positions = self.__get_log_files_initial_positions()
+
                 (
                     worker_thread,
                     self.__copying_manager,
                     self.__monitors_manager,
-                ) = start_worker_thread(self.__config, logs_initial_positions)
+                ) = start_worker_thread(
+                    self.__config, logs_initial_positions=logs_initial_positions
+                )
 
                 # NOTE: It's important we call this after worker thread has been created since
                 # some of the global configuration options are only applied after creating a worker
@@ -1354,11 +1439,20 @@ class ScalyrAgent(object):
                     prev_server = scalyr_server
 
                     log.info("Starting new copying and metrics threads")
+
+                    # get log files initial positions for new worker session log files if they were added in
+                    # a new configuration.
+                    logs_initial_positions = self.__get_log_files_initial_positions(
+                        only_new_files=True
+                    )
+
                     (
                         worker_thread,
                         self.__copying_manager,
                         self.__monitors_manager,
-                    ) = start_worker_thread(new_config)
+                    ) = start_worker_thread(
+                        new_config, logs_initial_positions=logs_initial_positions
+                    )
 
                     self.__current_bad_config = None
                     config_change_check_interval = (
